/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details. 
 *
 * octeon_soft_trace.c 
 *    octeon SMP software trace facility implementation
 *
 * Copyright (C) 2008 Cavium Networks
 */

#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/module.h>
#include <linux/percpu.h>
#include "octeon_soft_trace.h" 
#include "cvmx.h" 

#define ALLCORES -1
#define MODULE_NAME "octeon_soft_trace"

struct cvm_oct_trace_event {
         const char* filen;   /* File name */
	 const char* funcn;   /* Function name */
	 int line; 	/* Line number within function */
	 int core_id;   /* CPU core */
	 uint64_t ts;   /* global time stamp */
	 int num_p; 	/* Number of valid parameter */
	 int p[4]; 	/* Store up to 4 parameters */
};

/* One buffer per cpu,  
 *  buff_size and index are set up as per cpu data structure, to minize 
 *  locking and cache ping-pong. 
 */
struct cvm_oct_trace_buff {
	uint32_t processor_id;
	uint32_t buff_size;
	uint32_t index;
	uint32_t pr_index; /* print index, only used during output */
	struct cvm_oct_trace_event events[OCTEON_SOFT_TRACE_PERCPU_BUF_SIZE];
};

static struct cvm_oct_trace_buff *ev_buff;

/**
 * add trace event to trace buffer 
 *
 * @param  trace events
 * @return Zero on success
 */
int octeon_soft_trace_add_event  
	 (const char* filen, const char* funcn, int line,
	  int num_p, int p0, int p1, int p2, int p3)
{
#ifdef OCT_SOFT_TRACE_ENABLE
	int index, size; 
	struct cvm_oct_trace_buff *buff; /* per cpu pointer */
	struct cvm_oct_trace_event*  event;

	buff = percpu_ptr(ev_buff, smp_processor_id());

	index = buff->index;
	size  = buff->buff_size;
	event  = &(buff->events[index]);
	
	/* If we still have room */
	if (index < size - 1) {
		event->filen = filen;  
		event->funcn = funcn;  
		event->line = line;  
		event->core_id = cvmx_get_core_num();  
		event->ts = cvmx_get_cycle_global();  
		event->num_p = num_p;  
		event->p[0] = p0;  
		event->p[1] = p1;  
		event->p[2] = p2;  
		event->p[3] = p3;  

		buff->index++;
		return 0;
	}
#endif 

	return -1;
}
EXPORT_SYMBOL(octeon_soft_trace_add_event);

#ifdef OCT_SOFT_TRACE_ENABLE
/* Return the next event for output based on TS value; Increment Pindex
 * after the event is found. 
 */
static  struct cvm_oct_trace_event* get_next_event_for_output(int pr_id)
{
	uint64_t ts = -1;
	int cpu; 
	struct cvm_oct_trace_buff* buff = NULL;
	struct cvm_oct_trace_event* ev = NULL;
	int processor_id = pr_id;

	if (processor_id == ALLCORES) { 
		for_each_online_cpu(cpu) {
			struct cvm_oct_trace_buff* b;

			b = percpu_ptr(ev_buff, cpu);		
	        	if ( (b->pr_index < b->index) 
				&&  (b->events[b->pr_index].ts < ts) ) {
				buff = b;
				ts = b->events[b->pr_index].ts;
			}
		}
	}else { /* Per core output */
		struct cvm_oct_trace_buff* b;
		b = percpu_ptr(ev_buff, processor_id);		
	      	if (b->pr_index < b->index)  {
			buff = b;
		}
	}

	/* We have found the next event */
	if (buff) {
		ev = &buff->events[buff->pr_index];
		buff->pr_index++;
	}    

	return ev;
}

#endif

/* 
 *  Proc access to trace buffer 
 */

static int trace_buffer_read(struct seq_file *m, void *v ) 
{
#ifdef OCT_SOFT_TRACE_ENABLE
	int cpu, i, j, total; 
	int processor_id;
	struct cvm_oct_trace_event *cur = NULL, *last = NULL;  

	/* Assign processor_id  */
	if (m->private) {
		processor_id = (int) *(uint32_t *)(m->private);  /* Get processor it */
	}else {
		processor_id = ALLCORES;
	}
	
	if (processor_id == ALLCORES) { /* dump trace info for all cores */
		/* Print a summary of events per cpu and total events */
		total=0;	
		for_each_possible_cpu(cpu) {
	    	int count = percpu_ptr(ev_buff,cpu)->index;
	    	seq_printf (m, "%d:%d, ", cpu, count);
	    	total += count;
		}
		seq_printf (m, "total:%d\n" , total);


		/* Zero out the print index */
		for_each_possible_cpu(cpu) {
	    		percpu_ptr(ev_buff,cpu)->pr_index = 0;
		}
	}else {  
		/* Only dump from per cpu buffer */
	    	total  = percpu_ptr(ev_buff,processor_id)->index;
		percpu_ptr(ev_buff, processor_id)->pr_index = 0;
	}
	
	/* Output events in time stamp order */
	for (i = 0; i <  total; i++) {
		uint64_t clk_delta = 0; 

		last = cur;
		cur = get_next_event_for_output(processor_id);
		if (last)  {
			clk_delta =  cur->ts - last->ts;
			clk_delta &= 0xfffffffff; /* Only allow valid digits */
		}

		if (cur) {
			seq_printf (m, "%-5d %-2d: %30s:%4i %10lu %10lu  ", i, 
				cur->core_id,
				cur->funcn,
				cur->line,
				clk_delta, 
				cur->ts); 
		
			for ( j = 0; j< cur->num_p; j++) {
		    		seq_printf (m, "%5d  ", cur->p[j]); 
			}

			seq_printf (m, "\n" );
		}else {
			seq_printf (m, 
				"failed in getting the next event %d\n", i );
			break;
		}
        }
#else 
        seq_printf (m, "OCT_SOFT_TRACE_ENABLE is not defined , reported by %s, %s  \n ", __FILE__, __FUNCTION__); 
#endif 

	return 0;
}
/**
 *  /proc/octeon_soft_trace was openned. Use the single_open iterator
 *  
 *  @param inode
 *  @param file
 *  @return
 * */
static int trace_buffer_open(struct inode *inode, struct file *file)
{
        return single_open(file, trace_buffer_read, PDE(inode)->data);
}


static ssize_t trace_buffer_write(struct file *file, const char __user *buf, 
				  size_t count, loff_t *ppos) 
{
#ifdef OCT_SOFT_TRACE_ENABLE
	int cpu;

	/* zero trace buffer index on any write */
	for_each_possible_cpu(cpu) {
	    struct cvm_oct_trace_buff* buff = percpu_ptr(ev_buff, cpu);
	    uint64_t old_val;

	    old_val = buff->index;
            buff->index = 0;
	}
	fast_mb();
#endif 
	return count;

}

static const struct file_operations  trace_buffer_proc_operations = {
	.open     = trace_buffer_open, 
	.read     = seq_read, 
        .llseek   = seq_lseek,
	.write    = trace_buffer_write,
        .release  = single_release,
};

static int trace_buffer_proc_initialize(void) 
{
	struct proc_dir_entry *pr, *pe; 
	int cpu;
	int enabled;

	pr = proc_mkdir(MODULE_NAME,  NULL);
	if ( pr == NULL ) return  -ENOMEM;

#ifdef OCT_SOFT_TRACE_ENABLE
        pe = create_proc_entry("allcores", S_IFREG | S_IRUGO, pr);
	enabled = 1;
#else
        pe = create_proc_entry("notenabled", S_IFREG | S_IRUGO, pr);
	enabled = 0;
#endif
	
	if (pe) {
        	pe->proc_fops = &trace_buffer_proc_operations;
		pe->data =  NULL; /* AllCORES */
	} 

	if (enabled) for_each_possible_cpu(cpu) 
	{ char cname[100]; 
		sprintf (cname, "core%d", cpu);	

        	pe = create_proc_entry(cname, S_IFREG | S_IRUGO, pr);
		if (pe) {
			struct cvm_oct_trace_buff *buff; 
        		pe->proc_fops = &trace_buffer_proc_operations;			
			buff = percpu_ptr(ev_buff,cpu);
			pe->data = (void*) &(buff -> processor_id) ;
			pe->owner=THIS_MODULE;
		}		
	}

	return 0;
}

void trace_buffer_proc_shutdown(void) 
{

	remove_proc_entry(MODULE_NAME, NULL);
}

/**
 * Module initialization
 *
 * @return
 */
static int __init octeon_soft_trace_init(void)
{
#ifdef OCT_SOFT_TRACE_ENABLE
	int cpu;
	
	ev_buff =  alloc_percpu(struct cvm_oct_trace_buff); 
	if (ev_buff == NULL) 
		panic ("unable to allocate octeon soft trace buffer \n"); 

	/* Zero out the print index */
	for_each_possible_cpu(cpu) {
		struct cvm_oct_trace_buff *buff; 
        	buff = percpu_ptr(ev_buff,cpu);
		buff->processor_id = cpu;
   		buff->index = 0;
        	buff->pr_index = 0;
        	buff->buff_size = OCTEON_SOFT_TRACE_PERCPU_BUF_SIZE;
    	}
    	fast_mb();
#else 
    ev_buff = NULL;
#endif 
    trace_buffer_proc_initialize();

    return 0;
}

static void __exit  octeon_soft_trace_cleanup(void) 
{
    if (ev_buff) {
	free_percpu(ev_buff);
    }

    trace_buffer_proc_shutdown();
    
}


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Cavium Networks <support@caviumnetworks.com>");
MODULE_DESCRIPTION("Cavium Networks Octeon information interface.");
module_init(octeon_soft_trace_init);
module_exit(octeon_soft_trace_cleanup);
